@@ -19,20 +19,24 @@ import (
 	"net/url"
 	"path/filepath"
 	"regexp"
-	"strings"
 	"testing"
 	"time"
 
-	config_util "github.com/prometheus/common/config"
+	"github.com/prometheus/common/config"
 	"github.com/prometheus/common/model"
+	"github.com/stretchr/testify/require"
 	"gopkg.in/yaml.v2"
 
+	"github.com/prometheus/prometheus/discovery"
 	"github.com/prometheus/prometheus/discovery/azure"
-	sd_config "github.com/prometheus/prometheus/discovery/config"
 	"github.com/prometheus/prometheus/discovery/consul"
+	"github.com/prometheus/prometheus/discovery/digitalocean"
 	"github.com/prometheus/prometheus/discovery/dns"
+	"github.com/prometheus/prometheus/discovery/dockerswarm"
 	"github.com/prometheus/prometheus/discovery/ec2"
+	"github.com/prometheus/prometheus/discovery/eureka"
 	"github.com/prometheus/prometheus/discovery/file"
+	"github.com/prometheus/prometheus/discovery/hetzner"
 	"github.com/prometheus/prometheus/discovery/kubernetes"
 	"github.com/prometheus/prometheus/discovery/marathon"
 	"github.com/prometheus/prometheus/discovery/openstack"
@@ -41,15 +45,14 @@ import (
 	"github.com/prometheus/prometheus/discovery/zookeeper"
 	"github.com/prometheus/prometheus/pkg/labels"
 	"github.com/prometheus/prometheus/pkg/relabel"
-	"github.com/prometheus/prometheus/util/testutil"
 )
 
-func mustParseURL(u string) *config_util.URL {
+func mustParseURL(u string) *config.URL {
 	parsed, err := url.Parse(u)
 	if err != nil {
 		panic(err)
 	}
-	return &config_util.URL{URL: parsed}
+	return &config.URL{URL: parsed}
 }
 
 var expectedConf = &Config{
@@ -57,6 +60,7 @@ var expectedConf = &Config{
 		ScrapeInterval:     model.Duration(15 * time.Second),
 		ScrapeTimeout:      DefaultGlobalConfig.ScrapeTimeout,
 		EvaluationInterval: model.Duration(30 * time.Second),
+		QueryLogFile:       "",
 
 		ExternalLabels: labels.Labels{
 			{Name: "foo", Value: "bar"},
@@ -73,6 +77,7 @@ var expectedConf = &Config{
 		{
 			URL:           mustParseURL("http://remote1/push"),
 			RemoteTimeout: model.Duration(30 * time.Second),
+			Name:          "drop_expensive",
 			WriteRelabelConfigs: []*relabel.Config{
 				{
 					SourceLabels: model.LabelNames{"__name__"},
@@ -82,18 +87,22 @@ var expectedConf = &Config{
 					Action:       relabel.Drop,
 				},
 			},
-			QueueConfig: DefaultQueueConfig,
+			QueueConfig:    DefaultQueueConfig,
+			MetadataConfig: DefaultMetadataConfig,
 		},
 		{
-			URL:           mustParseURL("http://remote2/push"),
-			RemoteTimeout: model.Duration(30 * time.Second),
-			QueueConfig:   DefaultQueueConfig,
-			HTTPClientConfig: config_util.HTTPClientConfig{
-				TLSConfig: config_util.TLSConfig{
+			URL:            mustParseURL("http://remote2/push"),
+			RemoteTimeout:  model.Duration(30 * time.Second),
+			QueueConfig:    DefaultQueueConfig,
+			MetadataConfig: DefaultMetadataConfig,
+			Name:           "rw_tls",
+			HTTPClientConfig: config.HTTPClientConfig{
+				TLSConfig: config.TLSConfig{
 					CertFile: filepath.FromSlash("testdata/valid_cert_file"),
 					KeyFile:  filepath.FromSlash("testdata/valid_key_file"),
 				},
 			},
+			Headers: map[string]string{"name": "value"},
 		},
 	},
 
@@ -102,14 +111,16 @@ var expectedConf = &Config{
 			URL:           mustParseURL("http://remote1/read"),
 			RemoteTimeout: model.Duration(1 * time.Minute),
 			ReadRecent:    true,
+			Name:          "default",
 		},
 		{
 			URL:              mustParseURL("http://remote3/read"),
 			RemoteTimeout:    model.Duration(1 * time.Minute),
 			ReadRecent:       false,
+			Name:             "read_special",
 			RequiredMatchers: model.LabelSet{"job": "special"},
-			HTTPClientConfig: config_util.HTTPClientConfig{
-				TLSConfig: config_util.TLSConfig{
+			HTTPClientConfig: config.HTTPClientConfig{
+				TLSConfig: config.TLSConfig{
 					CertFile: filepath.FromSlash("testdata/valid_cert_file"),
 					KeyFile:  filepath.FromSlash("testdata/valid_key_file"),
 				},
@@ -129,12 +140,20 @@ var expectedConf = &Config{
 			MetricsPath: DefaultScrapeConfig.MetricsPath,
 			Scheme:      DefaultScrapeConfig.Scheme,
 
-			HTTPClientConfig: config_util.HTTPClientConfig{
+			HTTPClientConfig: config.HTTPClientConfig{
 				BearerTokenFile: filepath.FromSlash("testdata/valid_token_file"),
 			},
 
-			ServiceDiscoveryConfig: sd_config.ServiceDiscoveryConfig{
-				StaticConfigs: []*targetgroup.Group{
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&file.SDConfig{
+					Files:           []string{"testdata/foo/*.slow.json", "testdata/foo/*.slow.yml", "testdata/single/file.yml"},
+					RefreshInterval: model.Duration(10 * time.Minute),
+				},
+				&file.SDConfig{
+					Files:           []string{"testdata/bar/*.yaml"},
+					RefreshInterval: model.Duration(5 * time.Minute),
+				},
+				discovery.StaticConfig{
 					{
 						Targets: []model.LabelSet{
 							{model.AddressLabel: "localhost:9090"},
@@ -147,17 +166,6 @@ var expectedConf = &Config{
 						Source: "0",
 					},
 				},
-
-				FileSDConfigs: []*file.SDConfig{
-					{
-						Files:           []string{"testdata/foo/*.slow.json", "testdata/foo/*.slow.yml", "testdata/single/file.yml"},
-						RefreshInterval: model.Duration(10 * time.Minute),
-					},
-					{
-						Files:           []string{"testdata/bar/*.yaml"},
-						RefreshInterval: model.Duration(5 * time.Minute),
-					},
-				},
 			},
 
 			RelabelConfigs: []*relabel.Config{
@@ -199,8 +207,8 @@ var expectedConf = &Config{
 			ScrapeTimeout:   model.Duration(5 * time.Second),
 			SampleLimit:     1000,
 
-			HTTPClientConfig: config_util.HTTPClientConfig{
-				BasicAuth: &config_util.BasicAuth{
+			HTTPClientConfig: config.HTTPClientConfig{
+				BasicAuth: &config.BasicAuth{
 					Username: "admin_name",
 					Password: "multiline\nmysecret\ntest",
 				},
@@ -208,23 +216,21 @@ var expectedConf = &Config{
 			MetricsPath: "/my_path",
 			Scheme:      "https",
 
-			ServiceDiscoveryConfig: sd_config.ServiceDiscoveryConfig{
-				DNSSDConfigs: []*dns.SDConfig{
-					{
-						Names: []string{
-							"first.dns.address.domain.com",
-							"second.dns.address.domain.com",
-						},
-						RefreshInterval: model.Duration(15 * time.Second),
-						Type:            "SRV",
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&dns.SDConfig{
+					Names: []string{
+						"first.dns.address.domain.com",
+						"second.dns.address.domain.com",
 					},
-					{
-						Names: []string{
-							"first.dns.address.domain.com",
-						},
-						RefreshInterval: model.Duration(30 * time.Second),
-						Type:            "SRV",
+					RefreshInterval: model.Duration(15 * time.Second),
+					Type:            "SRV",
+				},
+				&dns.SDConfig{
+					Names: []string{
+						"first.dns.address.domain.com",
 					},
+					RefreshInterval: model.Duration(30 * time.Second),
+					Type:            "SRV",
 				},
 			},
 
@@ -291,24 +297,22 @@ var expectedConf = &Config{
 			MetricsPath: DefaultScrapeConfig.MetricsPath,
 			Scheme:      DefaultScrapeConfig.Scheme,
 
-			ServiceDiscoveryConfig: sd_config.ServiceDiscoveryConfig{
-				ConsulSDConfigs: []*consul.SDConfig{
-					{
-						Server:          "localhost:1234",
-						Token:           "mysecret",
-						Services:        []string{"nginx", "cache", "mysql"},
-						ServiceTags:     []string{"canary", "v1"},
-						NodeMeta:        map[string]string{"rack": "123"},
-						TagSeparator:    consul.DefaultSDConfig.TagSeparator,
-						Scheme:          "https",
-						RefreshInterval: consul.DefaultSDConfig.RefreshInterval,
-						AllowStale:      true,
-						TLSConfig: config_util.TLSConfig{
-							CertFile:           filepath.FromSlash("testdata/valid_cert_file"),
-							KeyFile:            filepath.FromSlash("testdata/valid_key_file"),
-							CAFile:             filepath.FromSlash("testdata/valid_ca_file"),
-							InsecureSkipVerify: false,
-						},
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&consul.SDConfig{
+					Server:          "localhost:1234",
+					Token:           "mysecret",
+					Services:        []string{"nginx", "cache", "mysql"},
+					ServiceTags:     []string{"canary", "v1"},
+					NodeMeta:        map[string]string{"rack": "123"},
+					TagSeparator:    consul.DefaultSDConfig.TagSeparator,
+					Scheme:          "https",
+					RefreshInterval: consul.DefaultSDConfig.RefreshInterval,
+					AllowStale:      true,
+					TLSConfig: config.TLSConfig{
+						CertFile:           filepath.FromSlash("testdata/valid_cert_file"),
+						KeyFile:            filepath.FromSlash("testdata/valid_key_file"),
+						CAFile:             filepath.FromSlash("testdata/valid_ca_file"),
+						InsecureSkipVerify: false,
 					},
 				},
 			},
@@ -334,8 +338,8 @@ var expectedConf = &Config{
 			MetricsPath: "/metrics",
 			Scheme:      "http",
 
-			HTTPClientConfig: config_util.HTTPClientConfig{
-				TLSConfig: config_util.TLSConfig{
+			HTTPClientConfig: config.HTTPClientConfig{
+				TLSConfig: config.TLSConfig{
 					CertFile: filepath.FromSlash("testdata/valid_cert_file"),
 					KeyFile:  filepath.FromSlash("testdata/valid_key_file"),
 				},
@@ -353,23 +357,21 @@ var expectedConf = &Config{
 			MetricsPath: DefaultScrapeConfig.MetricsPath,
 			Scheme:      DefaultScrapeConfig.Scheme,
 
-			ServiceDiscoveryConfig: sd_config.ServiceDiscoveryConfig{
-				KubernetesSDConfigs: []*kubernetes.SDConfig{
-					{
-						APIServer: kubernetesSDHostURL(),
-						Role:      kubernetes.RoleEndpoint,
-						HTTPClientConfig: config_util.HTTPClientConfig{
-							BasicAuth: &config_util.BasicAuth{
-								Username: "myusername",
-								Password: "mysecret",
-							},
-							TLSConfig: config_util.TLSConfig{
-								CertFile: filepath.FromSlash("testdata/valid_cert_file"),
-								KeyFile:  filepath.FromSlash("testdata/valid_key_file"),
-							},
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&kubernetes.SDConfig{
+					APIServer: kubernetesSDHostURL(),
+					Role:      kubernetes.RoleEndpoint,
+					HTTPClientConfig: config.HTTPClientConfig{
+						BasicAuth: &config.BasicAuth{
+							Username: "myusername",
+							Password: "mysecret",
+						},
+						TLSConfig: config.TLSConfig{
+							CertFile: filepath.FromSlash("testdata/valid_cert_file"),
+							KeyFile:  filepath.FromSlash("testdata/valid_key_file"),
 						},
-						NamespaceDiscovery: kubernetes.NamespaceDiscovery{},
 					},
+					NamespaceDiscovery: kubernetes.NamespaceDiscovery{},
 				},
 			},
 		},
@@ -382,22 +384,20 @@ var expectedConf = &Config{
 
 			MetricsPath: DefaultScrapeConfig.MetricsPath,
 			Scheme:      DefaultScrapeConfig.Scheme,
-			HTTPClientConfig: config_util.HTTPClientConfig{
-				BasicAuth: &config_util.BasicAuth{
+			HTTPClientConfig: config.HTTPClientConfig{
+				BasicAuth: &config.BasicAuth{
 					Username:     "myusername",
 					PasswordFile: filepath.FromSlash("testdata/valid_password_file"),
 				},
 			},
 
-			ServiceDiscoveryConfig: sd_config.ServiceDiscoveryConfig{
-				KubernetesSDConfigs: []*kubernetes.SDConfig{
-					{
-						APIServer: kubernetesSDHostURL(),
-						Role:      kubernetes.RoleEndpoint,
-						NamespaceDiscovery: kubernetes.NamespaceDiscovery{
-							Names: []string{
-								"default",
-							},
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&kubernetes.SDConfig{
+					APIServer: kubernetesSDHostURL(),
+					Role:      kubernetes.RoleEndpoint,
+					NamespaceDiscovery: kubernetes.NamespaceDiscovery{
+						Names: []string{
+							"default",
 						},
 					},
 				},
@@ -413,19 +413,17 @@ var expectedConf = &Config{
 			MetricsPath: DefaultScrapeConfig.MetricsPath,
 			Scheme:      DefaultScrapeConfig.Scheme,
 
-			ServiceDiscoveryConfig: sd_config.ServiceDiscoveryConfig{
-				MarathonSDConfigs: []*marathon.SDConfig{
-					{
-						Servers: []string{
-							"https://marathon.example.com:443",
-						},
-						RefreshInterval: model.Duration(30 * time.Second),
-						AuthToken:       config_util.Secret("mysecret"),
-						HTTPClientConfig: config_util.HTTPClientConfig{
-							TLSConfig: config_util.TLSConfig{
-								CertFile: filepath.FromSlash("testdata/valid_cert_file"),
-								KeyFile:  filepath.FromSlash("testdata/valid_key_file"),
-							},
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&marathon.SDConfig{
+					Servers: []string{
+						"https://marathon.example.com:443",
+					},
+					RefreshInterval: model.Duration(30 * time.Second),
+					AuthToken:       "mysecret",
+					HTTPClientConfig: config.HTTPClientConfig{
+						TLSConfig: config.TLSConfig{
+							CertFile: filepath.FromSlash("testdata/valid_cert_file"),
+							KeyFile:  filepath.FromSlash("testdata/valid_key_file"),
 						},
 					},
 				},
@@ -441,24 +439,22 @@ var expectedConf = &Config{
 			MetricsPath: DefaultScrapeConfig.MetricsPath,
 			Scheme:      DefaultScrapeConfig.Scheme,
 
-			ServiceDiscoveryConfig: sd_config.ServiceDiscoveryConfig{
-				EC2SDConfigs: []*ec2.SDConfig{
-					{
-						Region:          "us-east-1",
-						AccessKey:       "access",
-						SecretKey:       "mysecret",
-						Profile:         "profile",
-						RefreshInterval: model.Duration(60 * time.Second),
-						Port:            80,
-						Filters: []*ec2.Filter{
-							{
-								Name:   "tag:environment",
-								Values: []string{"prod"},
-							},
-							{
-								Name:   "tag:service",
-								Values: []string{"web", "db"},
-							},
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&ec2.SDConfig{
+					Region:          "us-east-1",
+					AccessKey:       "access",
+					SecretKey:       "mysecret",
+					Profile:         "profile",
+					RefreshInterval: model.Duration(60 * time.Second),
+					Port:            80,
+					Filters: []*ec2.Filter{
+						{
+							Name:   "tag:environment",
+							Values: []string{"prod"},
+						},
+						{
+							Name:   "tag:service",
+							Values: []string{"web", "db"},
 						},
 					},
 				},
@@ -474,18 +470,16 @@ var expectedConf = &Config{
 			MetricsPath: DefaultScrapeConfig.MetricsPath,
 			Scheme:      DefaultScrapeConfig.Scheme,
 
-			ServiceDiscoveryConfig: sd_config.ServiceDiscoveryConfig{
-				AzureSDConfigs: []*azure.SDConfig{
-					{
-						Environment:          "AzurePublicCloud",
-						SubscriptionID:       "11AAAA11-A11A-111A-A111-1111A1111A11",
-						TenantID:             "BBBB222B-B2B2-2B22-B222-2BB2222BB2B2",
-						ClientID:             "333333CC-3C33-3333-CCC3-33C3CCCCC33C",
-						ClientSecret:         "mysecret",
-						AuthenticationMethod: "OAuth",
-						RefreshInterval:      model.Duration(5 * time.Minute),
-						Port:                 9100,
-					},
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&azure.SDConfig{
+					Environment:          "AzurePublicCloud",
+					SubscriptionID:       "11AAAA11-A11A-111A-A111-1111A1111A11",
+					TenantID:             "BBBB222B-B2B2-2B22-B222-2BB2222BB2B2",
+					ClientID:             "333333CC-3C33-3333-CCC3-33C3CCCCC33C",
+					ClientSecret:         "mysecret",
+					AuthenticationMethod: "OAuth",
+					RefreshInterval:      model.Duration(5 * time.Minute),
+					Port:                 9100,
 				},
 			},
 		},
@@ -499,13 +493,11 @@ var expectedConf = &Config{
 			MetricsPath: DefaultScrapeConfig.MetricsPath,
 			Scheme:      DefaultScrapeConfig.Scheme,
 
-			ServiceDiscoveryConfig: sd_config.ServiceDiscoveryConfig{
-				NerveSDConfigs: []*zookeeper.NerveSDConfig{
-					{
-						Servers: []string{"localhost"},
-						Paths:   []string{"/monitoring"},
-						Timeout: model.Duration(10 * time.Second),
-					},
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&zookeeper.NerveSDConfig{
+					Servers: []string{"localhost"},
+					Paths:   []string{"/monitoring"},
+					Timeout: model.Duration(10 * time.Second),
 				},
 			},
 		},
@@ -519,8 +511,8 @@ var expectedConf = &Config{
 			MetricsPath: DefaultScrapeConfig.MetricsPath,
 			Scheme:      DefaultScrapeConfig.Scheme,
 
-			ServiceDiscoveryConfig: sd_config.ServiceDiscoveryConfig{
-				StaticConfigs: []*targetgroup.Group{
+			ServiceDiscoveryConfigs: discovery.Configs{
+				discovery.StaticConfig{
 					{
 						Targets: []model.LabelSet{
 							{model.AddressLabel: "localhost:9090"},
@@ -540,8 +532,8 @@ var expectedConf = &Config{
 			MetricsPath: "/federate",
 			Scheme:      DefaultScrapeConfig.Scheme,
 
-			ServiceDiscoveryConfig: sd_config.ServiceDiscoveryConfig{
-				StaticConfigs: []*targetgroup.Group{
+			ServiceDiscoveryConfigs: discovery.Configs{
+				discovery.StaticConfig{
 					{
 						Targets: []model.LabelSet{
 							{model.AddressLabel: "localhost:9090"},
@@ -561,8 +553,8 @@ var expectedConf = &Config{
 			MetricsPath: DefaultScrapeConfig.MetricsPath,
 			Scheme:      DefaultScrapeConfig.Scheme,
 
-			ServiceDiscoveryConfig: sd_config.ServiceDiscoveryConfig{
-				StaticConfigs: []*targetgroup.Group{
+			ServiceDiscoveryConfigs: discovery.Configs{
+				discovery.StaticConfig{
 					{
 						Targets: []model.LabelSet{
 							{model.AddressLabel: "localhost:9090"},
@@ -582,24 +574,62 @@ var expectedConf = &Config{
 			MetricsPath: DefaultScrapeConfig.MetricsPath,
 			Scheme:      DefaultScrapeConfig.Scheme,
 
-			ServiceDiscoveryConfig: sd_config.ServiceDiscoveryConfig{
-				TritonSDConfigs: []*triton.SDConfig{
-					{
-
-						Account:         "testAccount",
-						DNSSuffix:       "triton.example.com",
-						Endpoint:        "triton.example.com",
-						Port:            9163,
-						RefreshInterval: model.Duration(60 * time.Second),
-						Version:         1,
-						TLSConfig: config_util.TLSConfig{
-							CertFile: "testdata/valid_cert_file",
-							KeyFile:  "testdata/valid_key_file",
-						},
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&triton.SDConfig{
+					Account:         "testAccount",
+					Role:            "container",
+					DNSSuffix:       "triton.example.com",
+					Endpoint:        "triton.example.com",
+					Port:            9163,
+					RefreshInterval: model.Duration(60 * time.Second),
+					Version:         1,
+					TLSConfig: config.TLSConfig{
+						CertFile: "testdata/valid_cert_file",
+						KeyFile:  "testdata/valid_key_file",
 					},
 				},
 			},
 		},
+		{
+			JobName: "digitalocean-droplets",
+
+			HonorTimestamps: true,
+			ScrapeInterval:  model.Duration(15 * time.Second),
+			ScrapeTimeout:   DefaultGlobalConfig.ScrapeTimeout,
+
+			MetricsPath: DefaultScrapeConfig.MetricsPath,
+			Scheme:      DefaultScrapeConfig.Scheme,
+
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&digitalocean.SDConfig{
+					HTTPClientConfig: config.HTTPClientConfig{
+						BearerToken: "abcdef",
+					},
+					Port:            80,
+					RefreshInterval: model.Duration(60 * time.Second),
+				},
+			},
+		},
+		{
+			JobName: "dockerswarm",
+
+			HonorTimestamps: true,
+			ScrapeInterval:  model.Duration(15 * time.Second),
+			ScrapeTimeout:   DefaultGlobalConfig.ScrapeTimeout,
+
+			MetricsPath: DefaultScrapeConfig.MetricsPath,
+			Scheme:      DefaultScrapeConfig.Scheme,
+
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&dockerswarm.SDConfig{
+					Filters:         []dockerswarm.Filter{},
+					Host:            "http://127.0.0.1:2375",
+					Role:            "nodes",
+					Port:            80,
+					RefreshInterval: model.Duration(60 * time.Second),
+				},
+			},
+		},
 		{
 			JobName: "service-openstack",
 
@@ -610,20 +640,61 @@ var expectedConf = &Config{
 			MetricsPath: DefaultScrapeConfig.MetricsPath,
 			Scheme:      DefaultScrapeConfig.Scheme,
 
-			ServiceDiscoveryConfig: sd_config.ServiceDiscoveryConfig{
-				OpenstackSDConfigs: []*openstack.SDConfig{
-					{
-						Role:            "instance",
-						Region:          "RegionOne",
-						Port:            80,
-						RefreshInterval: model.Duration(60 * time.Second),
-						TLSConfig: config_util.TLSConfig{
-							CAFile:   "testdata/valid_ca_file",
-							CertFile: "testdata/valid_cert_file",
-							KeyFile:  "testdata/valid_key_file",
-						},
+			ServiceDiscoveryConfigs: discovery.Configs{&openstack.SDConfig{
+				Role:            "instance",
+				Region:          "RegionOne",
+				Port:            80,
+				Availability:    "public",
+				RefreshInterval: model.Duration(60 * time.Second),
+				TLSConfig: config.TLSConfig{
+					CAFile:   "testdata/valid_ca_file",
+					CertFile: "testdata/valid_cert_file",
+					KeyFile:  "testdata/valid_key_file",
+				}},
+			},
+		},
+		{
+			JobName:         "hetzner",
+			HonorTimestamps: true,
+			ScrapeInterval:  model.Duration(15 * time.Second),
+			ScrapeTimeout:   DefaultGlobalConfig.ScrapeTimeout,
+
+			MetricsPath: DefaultScrapeConfig.MetricsPath,
+			Scheme:      DefaultScrapeConfig.Scheme,
+
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&hetzner.SDConfig{
+					HTTPClientConfig: config.HTTPClientConfig{
+						BearerToken: "abcdef",
 					},
+					Port:            80,
+					RefreshInterval: model.Duration(60 * time.Second),
+					Role:            "hcloud",
 				},
+				&hetzner.SDConfig{
+					HTTPClientConfig: config.HTTPClientConfig{
+						BasicAuth: &config.BasicAuth{Username: "abcdef", Password: "abcdef"},
+					},
+					Port:            80,
+					RefreshInterval: model.Duration(60 * time.Second),
+					Role:            "robot",
+				},
+			},
+		},
+		{
+			JobName: "service-eureka",
+
+			HonorTimestamps: true,
+			ScrapeInterval:  model.Duration(15 * time.Second),
+			ScrapeTimeout:   DefaultGlobalConfig.ScrapeTimeout,
+
+			MetricsPath: DefaultScrapeConfig.MetricsPath,
+			Scheme:      DefaultScrapeConfig.Scheme,
+
+			ServiceDiscoveryConfigs: discovery.Configs{&eureka.SDConfig{
+				Server:          "http://eureka.example.com:8761/eureka",
+				RefreshInterval: model.Duration(30 * time.Second),
+			},
 			},
 		},
 	},
@@ -633,8 +704,8 @@ var expectedConf = &Config{
 				Scheme:     "https",
 				Timeout:    model.Duration(10 * time.Second),
 				APIVersion: AlertmanagerAPIVersionV1,
-				ServiceDiscoveryConfig: sd_config.ServiceDiscoveryConfig{
-					StaticConfigs: []*targetgroup.Group{
+				ServiceDiscoveryConfigs: discovery.Configs{
+					discovery.StaticConfig{
 						{
 							Targets: []model.LabelSet{
 								{model.AddressLabel: "1.2.3.4:9093"},
@@ -648,51 +719,81 @@ var expectedConf = &Config{
 			},
 		},
 	},
-	original: "",
+}
+
+func TestYAMLRoundtrip(t *testing.T) {
+	want, err := LoadFile("testdata/roundtrip.good.yml")
+	require.NoError(t, err)
+
+	out, err := yaml.Marshal(want)
+
+	require.NoError(t, err)
+	got := &Config{}
+	require.NoError(t, yaml.UnmarshalStrict(out, got))
+
+	require.Equal(t, want, got)
 }
 
 func TestLoadConfig(t *testing.T) {
 	// Parse a valid file that sets a global scrape timeout. This tests whether parsing
 	// an overwritten default field in the global config permanently changes the default.
 	_, err := LoadFile("testdata/global_timeout.good.yml")
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	c, err := LoadFile("testdata/conf.good.yml")
-	testutil.Ok(t, err)
+	require.NoError(t, err)
+	require.Equal(t, expectedConf, c)
+}
 
-	expectedConf.original = c.original
-	testutil.Equals(t, expectedConf, c)
+func TestScrapeIntervalLarger(t *testing.T) {
+	c, err := LoadFile("testdata/scrape_interval_larger.good.yml")
+	require.NoError(t, err)
+	require.Equal(t, 1, len(c.ScrapeConfigs))
+	for _, sc := range c.ScrapeConfigs {
+		require.Equal(t, true, sc.ScrapeInterval >= sc.ScrapeTimeout)
+	}
 }
 
 // YAML marshaling must not reveal authentication credentials.
 func TestElideSecrets(t *testing.T) {
 	c, err := LoadFile("testdata/conf.good.yml")
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	secretRe := regexp.MustCompile(`\\u003csecret\\u003e|<secret>`)
 
 	config, err := yaml.Marshal(c)
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 	yamlConfig := string(config)
 
 	matches := secretRe.FindAllStringIndex(yamlConfig, -1)
-	testutil.Assert(t, len(matches) == 7, "wrong number of secret matches found")
-	testutil.Assert(t, !strings.Contains(yamlConfig, "mysecret"),
+	require.Equal(t, 10, len(matches), "wrong number of secret matches found")
+	require.NotContains(t, yamlConfig, "mysecret",
 		"yaml marshal reveals authentication credentials.")
 }
 
 func TestLoadConfigRuleFilesAbsolutePath(t *testing.T) {
 	// Parse a valid file that sets a rule files with an absolute path
 	c, err := LoadFile(ruleFilesConfigFile)
-	testutil.Ok(t, err)
-
-	ruleFilesExpectedConf.original = c.original
-	testutil.Equals(t, ruleFilesExpectedConf, c)
+	require.NoError(t, err)
+	require.Equal(t, ruleFilesExpectedConf, c)
 }
 
 func TestKubernetesEmptyAPIServer(t *testing.T) {
 	_, err := LoadFile("testdata/kubernetes_empty_apiserver.good.yml")
-	testutil.Ok(t, err)
+	require.NoError(t, err)
+}
+
+func TestKubernetesSelectors(t *testing.T) {
+	_, err := LoadFile("testdata/kubernetes_selectors_endpoints.good.yml")
+	require.NoError(t, err)
+	_, err = LoadFile("testdata/kubernetes_selectors_node.good.yml")
+	require.NoError(t, err)
+	_, err = LoadFile("testdata/kubernetes_selectors_ingress.good.yml")
+	require.NoError(t, err)
+	_, err = LoadFile("testdata/kubernetes_selectors_pod.good.yml")
+	require.NoError(t, err)
+	_, err = LoadFile("testdata/kubernetes_selectors_service.good.yml")
+	require.NoError(t, err)
 }
 
 var expectedErrors = []struct {
@@ -761,7 +862,7 @@ var expectedErrors = []struct {
 		errMsg:   "invalid rule file path",
 	}, {
 		filename: "unknown_attr.bad.yml",
-		errMsg:   "field consult_sd_configs not found in type config.plain",
+		errMsg:   "field consult_sd_configs not found in type",
 	}, {
 		filename: "bearertoken.bad.yml",
 		errMsg:   "at most one of bearer_token & bearer_token_file must be configured",
@@ -777,9 +878,30 @@ var expectedErrors = []struct {
 	}, {
 		filename: "kubernetes_role.bad.yml",
 		errMsg:   "role",
+	}, {
+		filename: "kubernetes_selectors_endpoints.bad.yml",
+		errMsg:   "endpoints role supports only pod, service, endpoints selectors",
+	}, {
+		filename: "kubernetes_selectors_ingress.bad.yml",
+		errMsg:   "ingress role supports only ingress selectors",
+	}, {
+		filename: "kubernetes_selectors_node.bad.yml",
+		errMsg:   "node role supports only node selectors",
+	}, {
+		filename: "kubernetes_selectors_pod.bad.yml",
+		errMsg:   "pod role supports only pod selectors",
+	}, {
+		filename: "kubernetes_selectors_service.bad.yml",
+		errMsg:   "service role supports only service selectors",
 	}, {
 		filename: "kubernetes_namespace_discovery.bad.yml",
 		errMsg:   "field foo not found in type kubernetes.plain",
+	}, {
+		filename: "kubernetes_selectors_duplicated_role.bad.yml",
+		errMsg:   "duplicated selector role: pod",
+	}, {
+		filename: "kubernetes_selectors_incorrect_selector.bad.yml",
+		errMsg:   "invalid selector: 'metadata.status-Running'; can't understand 'metadata.status-Running'",
 	}, {
 		filename: "kubernetes_bearertoken_basicauth.bad.yml",
 		errMsg:   "at most one of basic_auth, bearer_token & bearer_token_file must be configured",
@@ -798,6 +920,9 @@ var expectedErrors = []struct {
 	}, {
 		filename: "openstack_role.bad.yml",
 		errMsg:   "unknown OpenStack SD role",
+	}, {
+		filename: "openstack_availability.bad.yml",
+		errMsg:   "unknown availability invalid, must be one of admin, internal or public",
 	}, {
 		filename: "url_in_targetgroup.bad.yml",
 		errMsg:   "\"http://bad\" is not a valid hostname",
@@ -813,9 +938,21 @@ var expectedErrors = []struct {
 	}, {
 		filename: "remote_read_url_missing.bad.yml",
 		errMsg:   `url for remote_read is empty`,
+	}, {
+		filename: "remote_write_header.bad.yml",
+		errMsg:   `x-prometheus-remote-write-version is an unchangeable header`,
+	}, {
+		filename: "remote_write_authorization_header.bad.yml",
+		errMsg:   `authorization header must be changed via the basic_auth, bearer_token, or bearer_token_file parameter`,
 	}, {
 		filename: "remote_write_url_missing.bad.yml",
 		errMsg:   `url for remote_write is empty`,
+	}, {
+		filename: "remote_write_dup.bad.yml",
+		errMsg:   `found multiple remote write configs with job name "queue1"`,
+	}, {
+		filename: "remote_read_dup.bad.yml",
+		errMsg:   `found multiple remote read configs with job name "queue1"`,
 	},
 	{
 		filename: "ec2_filters_empty_values.bad.yml",
@@ -881,49 +1018,60 @@ var expectedErrors = []struct {
 		filename: "empty_static_config.bad.yml",
 		errMsg:   "empty or null section in static_configs",
 	},
+	{
+		filename: "hetzner_role.bad.yml",
+		errMsg:   "unknown role",
+	},
+	{
+		filename: "eureka_no_server.bad.yml",
+		errMsg:   "empty or null eureka server",
+	},
+	{
+		filename: "eureka_invalid_server.bad.yml",
+		errMsg:   "invalid eureka server URL",
+	},
 }
 
 func TestBadConfigs(t *testing.T) {
 	for _, ee := range expectedErrors {
 		_, err := LoadFile("testdata/" + ee.filename)
-		testutil.NotOk(t, err, "%s", ee.filename)
-		testutil.Assert(t, strings.Contains(err.Error(), ee.errMsg),
+		require.Error(t, err, "%s", ee.filename)
+		require.Contains(t, err.Error(), ee.errMsg,
 			"Expected error for %s to contain %q but got: %s", ee.filename, ee.errMsg, err)
 	}
 }
 
 func TestBadStaticConfigsJSON(t *testing.T) {
 	content, err := ioutil.ReadFile("testdata/static_config.bad.json")
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 	var tg targetgroup.Group
 	err = json.Unmarshal(content, &tg)
-	testutil.NotOk(t, err, "")
+	require.Error(t, err)
 }
 
 func TestBadStaticConfigsYML(t *testing.T) {
 	content, err := ioutil.ReadFile("testdata/static_config.bad.yml")
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 	var tg targetgroup.Group
 	err = yaml.UnmarshalStrict(content, &tg)
-	testutil.NotOk(t, err, "")
+	require.Error(t, err)
 }
 
 func TestEmptyConfig(t *testing.T) {
 	c, err := Load("")
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 	exp := DefaultConfig
-	testutil.Equals(t, exp, *c)
+	require.Equal(t, exp, *c)
 }
 
 func TestEmptyGlobalBlock(t *testing.T) {
 	c, err := Load("global:\n")
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 	exp := DefaultConfig
-	exp.original = "global:\n"
-	testutil.Equals(t, exp, *c)
+	require.Equal(t, exp, *c)
 }
 
-func kubernetesSDHostURL() config_util.URL {
+func kubernetesSDHostURL() config.URL {
 	tURL, _ := url.Parse("https://localhost:1234")
-	return config_util.URL{URL: tURL}
+	return config.URL{URL: tURL}
 }
