@@ -14,10 +14,15 @@
 package adapter
 
 import (
+	"context"
+	"io/ioutil"
+	"os"
 	"reflect"
 	"testing"
 
 	"github.com/prometheus/common/model"
+	"github.com/stretchr/testify/require"
+
 	"github.com/prometheus/prometheus/discovery/targetgroup"
 )
 
@@ -222,3 +227,14 @@ func TestGenerateTargetGroups(t *testing.T) {
 
 	}
 }
+
+// TestWriteOutput checks the adapter can write a file to disk.
+func TestWriteOutput(t *testing.T) {
+	ctx := context.Background()
+	tmpfile, err := ioutil.TempFile("", "sd_adapter_test")
+	require.NoError(t, err)
+	defer os.Remove(tmpfile.Name())
+	tmpfile.Close()
+	adapter := NewAdapter(ctx, tmpfile.Name(), "test_sd", nil, nil)
+	require.NoError(t, adapter.writeOutput())
+}
