@@ -18,7 +18,6 @@ import (
 	"math"
 	"strconv"
 	"sync"
-	"sync/atomic"
 	"time"
 
 	"github.com/go-kit/kit/log"
@@ -27,12 +26,14 @@ import (
 	"github.com/golang/snappy"
 	"github.com/opentracing/opentracing-go"
 	"github.com/opentracing/opentracing-go/ext"
+	"go.uber.org/atomic"
 
 	"github.com/prometheus/client_golang/prometheus"
 	"github.com/prometheus/prometheus/config"
 	"github.com/prometheus/prometheus/pkg/labels"
 	"github.com/prometheus/prometheus/pkg/relabel"
 	"github.com/prometheus/prometheus/prompb"
+	"github.com/prometheus/prometheus/scrape"
 	"github.com/prometheus/prometheus/tsdb/record"
 	"github.com/prometheus/prometheus/tsdb/wal"
 )
@@ -50,20 +51,25 @@ const (
 type queueManagerMetrics struct {
 	reg prometheus.Registerer
 
-	succeededSamplesTotal prometheus.Counter
-	failedSamplesTotal    prometheus.Counter
-	retriedSamplesTotal   prometheus.Counter
-	droppedSamplesTotal   prometheus.Counter
-	enqueueRetriesTotal   prometheus.Counter
-	sentBatchDuration     prometheus.Histogram
-	highestSentTimestamp  *maxGauge
-	pendingSamples        prometheus.Gauge
-	shardCapacity         prometheus.Gauge
-	numShards             prometheus.Gauge
-	maxNumShards          prometheus.Gauge
-	minNumShards          prometheus.Gauge
-	desiredNumShards      prometheus.Gauge
-	bytesSent             prometheus.Counter
+	samplesTotal         prometheus.Counter
+	metadataTotal        prometheus.Counter
+	failedSamplesTotal   prometheus.Counter
+	failedMetadataTotal  prometheus.Counter
+	retriedSamplesTotal  prometheus.Counter
+	retriedMetadataTotal prometheus.Counter
+	droppedSamplesTotal  prometheus.Counter
+	enqueueRetriesTotal  prometheus.Counter
+	sentBatchDuration    prometheus.Histogram
+	highestSentTimestamp *maxTimestamp
+	pendingSamples       prometheus.Gauge
+	shardCapacity        prometheus.Gauge
+	numShards            prometheus.Gauge
+	maxNumShards         prometheus.Gauge
+	minNumShards         prometheus.Gauge
+	desiredNumShards     prometheus.Gauge
+	samplesBytesTotal    prometheus.Counter
+	metadataBytesTotal   prometheus.Counter
+	maxSamplesPerSend    prometheus.Gauge
 }
 
 func newQueueManagerMetrics(r prometheus.Registerer, rn, e string) *queueManagerMetrics {
@@ -75,31 +81,52 @@ func newQueueManagerMetrics(r prometheus.Registerer, rn, e string) *queueManager
 		endpoint:   e,
 	}
 
-	m.succeededSamplesTotal = prometheus.NewCounter(prometheus.CounterOpts{
+	m.samplesTotal = prometheus.NewCounter(prometheus.CounterOpts{
 		Namespace:   namespace,
 		Subsystem:   subsystem,
-		Name:        "succeeded_samples_total",
-		Help:        "Total number of samples successfully sent to remote storage.",
+		Name:        "samples_total",
+		Help:        "Total number of samples sent to remote storage.",
+		ConstLabels: constLabels,
+	})
+	m.metadataTotal = prometheus.NewCounter(prometheus.CounterOpts{
+		Namespace:   namespace,
+		Subsystem:   subsystem,
+		Name:        "metadata_total",
+		Help:        "Total number of metadata entries sent to remote storage.",
 		ConstLabels: constLabels,
 	})
 	m.failedSamplesTotal = prometheus.NewCounter(prometheus.CounterOpts{
 		Namespace:   namespace,
 		Subsystem:   subsystem,
-		Name:        "failed_samples_total",
+		Name:        "samples_failed_total",
 		Help:        "Total number of samples which failed on send to remote storage, non-recoverable errors.",
 		ConstLabels: constLabels,
 	})
+	m.failedMetadataTotal = prometheus.NewCounter(prometheus.CounterOpts{
+		Namespace:   namespace,
+		Subsystem:   subsystem,
+		Name:        "metadata_failed_total",
+		Help:        "Total number of metadata entries which failed on send to remote storage, non-recoverable errors.",
+		ConstLabels: constLabels,
+	})
 	m.retriedSamplesTotal = prometheus.NewCounter(prometheus.CounterOpts{
 		Namespace:   namespace,
 		Subsystem:   subsystem,
-		Name:        "retried_samples_total",
+		Name:        "samples_retried_total",
 		Help:        "Total number of samples which failed on send to remote storage but were retried because the send error was recoverable.",
 		ConstLabels: constLabels,
 	})
+	m.retriedMetadataTotal = prometheus.NewCounter(prometheus.CounterOpts{
+		Namespace:   namespace,
+		Subsystem:   subsystem,
+		Name:        "metadata_retried_total",
+		Help:        "Total number of metadata entries which failed on send to remote storage but were retried because the send error was recoverable.",
+		ConstLabels: constLabels,
+	})
 	m.droppedSamplesTotal = prometheus.NewCounter(prometheus.CounterOpts{
 		Namespace:   namespace,
 		Subsystem:   subsystem,
-		Name:        "dropped_samples_total",
+		Name:        "samples_dropped_total",
 		Help:        "Total number of samples which were dropped after being read from the WAL before being sent via remote write.",
 		ConstLabels: constLabels,
 	})
@@ -114,11 +141,11 @@ func newQueueManagerMetrics(r prometheus.Registerer, rn, e string) *queueManager
 		Namespace:   namespace,
 		Subsystem:   subsystem,
 		Name:        "sent_batch_duration_seconds",
-		Help:        "Duration of sample batch send calls to the remote storage.",
-		Buckets:     prometheus.DefBuckets,
+		Help:        "Duration of send calls to the remote storage.",
+		Buckets:     append(prometheus.DefBuckets, 25, 60, 120, 300),
 		ConstLabels: constLabels,
 	})
-	m.highestSentTimestamp = &maxGauge{
+	m.highestSentTimestamp = &maxTimestamp{
 		Gauge: prometheus.NewGauge(prometheus.GaugeOpts{
 			Namespace:   namespace,
 			Subsystem:   subsystem,
@@ -130,7 +157,7 @@ func newQueueManagerMetrics(r prometheus.Registerer, rn, e string) *queueManager
 	m.pendingSamples = prometheus.NewGauge(prometheus.GaugeOpts{
 		Namespace:   namespace,
 		Subsystem:   subsystem,
-		Name:        "pending_samples",
+		Name:        "samples_pending",
 		Help:        "The number of samples pending in the queues shards to be sent to the remote storage.",
 		ConstLabels: constLabels,
 	})
@@ -169,11 +196,25 @@ func newQueueManagerMetrics(r prometheus.Registerer, rn, e string) *queueManager
 		Help:        "The number of shards that the queues shard calculation wants to run based on the rate of samples in vs. samples out.",
 		ConstLabels: constLabels,
 	})
-	m.bytesSent = prometheus.NewCounter(prometheus.CounterOpts{
+	m.samplesBytesTotal = prometheus.NewCounter(prometheus.CounterOpts{
 		Namespace:   namespace,
 		Subsystem:   subsystem,
-		Name:        "sent_bytes_total",
-		Help:        "The total number of bytes sent by the queue.",
+		Name:        "samples_bytes_total",
+		Help:        "The total number of bytes of samples sent by the queue after compression.",
+		ConstLabels: constLabels,
+	})
+	m.metadataBytesTotal = prometheus.NewCounter(prometheus.CounterOpts{
+		Namespace:   namespace,
+		Subsystem:   subsystem,
+		Name:        "metadata_bytes_total",
+		Help:        "The total number of bytes of metadata sent by the queue after compression.",
+		ConstLabels: constLabels,
+	})
+	m.maxSamplesPerSend = prometheus.NewGauge(prometheus.GaugeOpts{
+		Namespace:   namespace,
+		Subsystem:   subsystem,
+		Name:        "max_samples_per_send",
+		Help:        "The maximum number of samples to be sent, in a single request, to the remote storage.",
 		ConstLabels: constLabels,
 	})
 
@@ -183,9 +224,12 @@ func newQueueManagerMetrics(r prometheus.Registerer, rn, e string) *queueManager
 func (m *queueManagerMetrics) register() {
 	if m.reg != nil {
 		m.reg.MustRegister(
-			m.succeededSamplesTotal,
+			m.samplesTotal,
+			m.metadataTotal,
 			m.failedSamplesTotal,
+			m.failedMetadataTotal,
 			m.retriedSamplesTotal,
+			m.retriedMetadataTotal,
 			m.droppedSamplesTotal,
 			m.enqueueRetriesTotal,
 			m.sentBatchDuration,
@@ -196,16 +240,21 @@ func (m *queueManagerMetrics) register() {
 			m.maxNumShards,
 			m.minNumShards,
 			m.desiredNumShards,
-			m.bytesSent,
+			m.samplesBytesTotal,
+			m.metadataBytesTotal,
+			m.maxSamplesPerSend,
 		)
 	}
 }
 
 func (m *queueManagerMetrics) unregister() {
 	if m.reg != nil {
-		m.reg.Unregister(m.succeededSamplesTotal)
+		m.reg.Unregister(m.samplesTotal)
+		m.reg.Unregister(m.metadataTotal)
 		m.reg.Unregister(m.failedSamplesTotal)
+		m.reg.Unregister(m.failedMetadataTotal)
 		m.reg.Unregister(m.retriedSamplesTotal)
+		m.reg.Unregister(m.retriedMetadataTotal)
 		m.reg.Unregister(m.droppedSamplesTotal)
 		m.reg.Unregister(m.enqueueRetriesTotal)
 		m.reg.Unregister(m.sentBatchDuration)
@@ -216,13 +265,15 @@ func (m *queueManagerMetrics) unregister() {
 		m.reg.Unregister(m.maxNumShards)
 		m.reg.Unregister(m.minNumShards)
 		m.reg.Unregister(m.desiredNumShards)
-		m.reg.Unregister(m.bytesSent)
+		m.reg.Unregister(m.samplesBytesTotal)
+		m.reg.Unregister(m.metadataBytesTotal)
+		m.reg.Unregister(m.maxSamplesPerSend)
 	}
 }
 
-// StorageClient defines an interface for sending a batch of samples to an
+// WriteClient defines an interface for sending a batch of samples to an
 // external timeseries database.
-type StorageClient interface {
+type WriteClient interface {
 	// Store stores the given samples in the remote storage.
 	Store(context.Context, []byte) error
 	// Name uniquely identifies the remote storage.
@@ -232,21 +283,22 @@ type StorageClient interface {
 }
 
 // QueueManager manages a queue of samples to be sent to the Storage
-// indicated by the provided StorageClient. Implements writeTo interface
+// indicated by the provided WriteClient. Implements writeTo interface
 // used by WAL Watcher.
 type QueueManager struct {
-	// https://golang.org/pkg/sync/atomic/#pkg-note-BUG
-	lastSendTimestamp int64
+	lastSendTimestamp atomic.Int64
 
-	logger         log.Logger
-	flushDeadline  time.Duration
-	cfg            config.QueueConfig
-	externalLabels labels.Labels
-	relabelConfigs []*relabel.Config
-	watcher        *wal.Watcher
+	logger          log.Logger
+	flushDeadline   time.Duration
+	cfg             config.QueueConfig
+	mcfg            config.MetadataConfig
+	externalLabels  labels.Labels
+	relabelConfigs  []*relabel.Config
+	watcher         *wal.Watcher
+	metadataWatcher *MetadataWatcher
 
 	clientMtx   sync.RWMutex
-	storeClient StorageClient
+	storeClient WriteClient
 
 	seriesMtx            sync.Mutex
 	seriesLabels         map[uint64]labels.Labels
@@ -261,7 +313,9 @@ type QueueManager struct {
 
 	samplesIn, samplesDropped, samplesOut, samplesOutDuration *ewmaRate
 
-	metrics *queueManagerMetrics
+	metrics              *queueManagerMetrics
+	interner             *pool
+	highestRecvTimestamp *maxTimestamp
 }
 
 // NewQueueManager builds a new QueueManager.
@@ -273,10 +327,14 @@ func NewQueueManager(
 	walDir string,
 	samplesIn *ewmaRate,
 	cfg config.QueueConfig,
+	mCfg config.MetadataConfig,
 	externalLabels labels.Labels,
 	relabelConfigs []*relabel.Config,
-	client StorageClient,
+	client WriteClient,
 	flushDeadline time.Duration,
+	interner *pool,
+	highestRecvTimestamp *maxTimestamp,
+	sm ReadyScrapeManager,
 ) *QueueManager {
 	if logger == nil {
 		logger = log.NewNopLogger()
@@ -287,6 +345,7 @@ func NewQueueManager(
 		logger:         logger,
 		flushDeadline:  flushDeadline,
 		cfg:            cfg,
+		mcfg:           mCfg,
 		externalLabels: externalLabels,
 		relabelConfigs: relabelConfigs,
 		storeClient:    client,
@@ -304,15 +363,83 @@ func NewQueueManager(
 		samplesOut:         newEWMARate(ewmaWeight, shardUpdateDuration),
 		samplesOutDuration: newEWMARate(ewmaWeight, shardUpdateDuration),
 
-		metrics: metrics,
+		metrics:              metrics,
+		interner:             interner,
+		highestRecvTimestamp: highestRecvTimestamp,
 	}
 
 	t.watcher = wal.NewWatcher(watcherMetrics, readerMetrics, logger, client.Name(), t, walDir)
+	if t.mcfg.Send {
+		t.metadataWatcher = NewMetadataWatcher(logger, sm, client.Name(), t, t.mcfg.SendInterval, flushDeadline)
+	}
 	t.shards = t.newShards()
 
 	return t
 }
 
+// AppendMetadata sends metadata the remote storage. Metadata is sent all at once and is not parallelized.
+func (t *QueueManager) AppendMetadata(ctx context.Context, metadata []scrape.MetricMetadata) {
+	mm := make([]prompb.MetricMetadata, 0, len(metadata))
+	for _, entry := range metadata {
+		mm = append(mm, prompb.MetricMetadata{
+			MetricFamilyName: entry.Metric,
+			Help:             entry.Help,
+			Type:             metricTypeToMetricTypeProto(entry.Type),
+			Unit:             entry.Unit,
+		})
+	}
+
+	err := t.sendMetadataWithBackoff(ctx, mm)
+
+	if err != nil {
+		t.metrics.failedMetadataTotal.Add(float64(len(metadata)))
+		level.Error(t.logger).Log("msg", "non-recoverable error while sending metadata", "count", len(metadata), "err", err)
+	}
+}
+
+func (t *QueueManager) sendMetadataWithBackoff(ctx context.Context, metadata []prompb.MetricMetadata) error {
+	// Build the WriteRequest with no samples.
+	req, _, err := buildWriteRequest(nil, metadata, nil)
+	if err != nil {
+		return err
+	}
+
+	metadataCount := len(metadata)
+
+	attemptStore := func(try int) error {
+		span, ctx := opentracing.StartSpanFromContext(ctx, "Remote Metadata Send Batch")
+		defer span.Finish()
+
+		span.SetTag("metadata", metadataCount)
+		span.SetTag("try", try)
+		span.SetTag("remote_name", t.storeClient.Name())
+		span.SetTag("remote_url", t.storeClient.Endpoint())
+
+		begin := time.Now()
+		err := t.storeClient.Store(ctx, req)
+		t.metrics.sentBatchDuration.Observe(time.Since(begin).Seconds())
+
+		if err != nil {
+			span.LogKV("error", err)
+			ext.Error.Set(span, true)
+			return err
+		}
+
+		return nil
+	}
+
+	retry := func() {
+		t.metrics.retriedMetadataTotal.Add(float64(len(metadata)))
+	}
+	err = sendWriteRequestWithBackoff(ctx, t.cfg, t.logger, attemptStore, retry)
+	if err != nil {
+		return err
+	}
+	t.metrics.metadataTotal.Add(float64(len(metadata)))
+	t.metrics.metadataBytesTotal.Add(float64(len(req)))
+	return nil
+}
+
 // Append queues a sample to be sent to the remote storage. Blocks until all samples are
 // enqueued on their shards or a shutdown signal is received.
 func (t *QueueManager) Append(samples []record.RefSample) bool {
@@ -364,13 +491,16 @@ func (t *QueueManager) Start() {
 	// Register and initialise some metrics.
 	t.metrics.register()
 	t.metrics.shardCapacity.Set(float64(t.cfg.Capacity))
-	t.metrics.pendingSamples.Set(0)
 	t.metrics.maxNumShards.Set(float64(t.cfg.MaxShards))
 	t.metrics.minNumShards.Set(float64(t.cfg.MinShards))
 	t.metrics.desiredNumShards.Set(float64(t.cfg.MinShards))
+	t.metrics.maxSamplesPerSend.Set(float64(t.cfg.MaxSamplesPerSend))
 
 	t.shards.start(t.numShards)
 	t.watcher.Start()
+	if t.mcfg.Send {
+		t.metadataWatcher.Start()
+	}
 
 	t.wg.Add(2)
 	go t.updateShardsLoop()
@@ -385,16 +515,19 @@ func (t *QueueManager) Stop() {
 
 	close(t.quit)
 	t.wg.Wait()
-	// Wait for all QueueManager routines to end before stopping shards and WAL watcher. This
+	// Wait for all QueueManager routines to end before stopping shards, metadata watcher, and WAL watcher. This
 	// is to ensure we don't end up executing a reshard and shards.stop() at the same time, which
 	// causes a closed channel panic.
 	t.shards.stop()
 	t.watcher.Stop()
+	if t.mcfg.Send {
+		t.metadataWatcher.Stop()
+	}
 
 	// On shutdown, release the strings in the labels from the intern pool.
 	t.seriesMtx.Lock()
 	for _, labels := range t.seriesLabels {
-		releaseLabels(labels)
+		t.releaseLabels(labels)
 	}
 	t.seriesMtx.Unlock()
 	t.metrics.unregister()
@@ -405,20 +538,22 @@ func (t *QueueManager) StoreSeries(series []record.RefSeries, index int) {
 	t.seriesMtx.Lock()
 	defer t.seriesMtx.Unlock()
 	for _, s := range series {
+		// Just make sure all the Refs of Series will insert into seriesSegmentIndexes map for tracking.
+		t.seriesSegmentIndexes[s.Ref] = index
+
 		ls := processExternalLabels(s.Labels, t.externalLabels)
 		lbls := relabel.Process(ls, t.relabelConfigs...)
 		if len(lbls) == 0 {
 			t.droppedSeries[s.Ref] = struct{}{}
 			continue
 		}
-		t.seriesSegmentIndexes[s.Ref] = index
-		internLabels(lbls)
+		t.internLabels(lbls)
 
 		// We should not ever be replacing a series labels in the map, but just
 		// in case we do we need to ensure we do not leak the replaced interned
 		// strings.
 		if orig, ok := t.seriesLabels[s.Ref]; ok {
-			releaseLabels(orig)
+			t.releaseLabels(orig)
 		}
 		t.seriesLabels[s.Ref] = lbls
 	}
@@ -435,7 +570,7 @@ func (t *QueueManager) SeriesReset(index int) {
 	for k, v := range t.seriesSegmentIndexes {
 		if v < index {
 			delete(t.seriesSegmentIndexes, k)
-			releaseLabels(t.seriesLabels[k])
+			t.releaseLabels(t.seriesLabels[k])
 			delete(t.seriesLabels, k)
 			delete(t.droppedSeries, k)
 		}
@@ -444,29 +579,29 @@ func (t *QueueManager) SeriesReset(index int) {
 
 // SetClient updates the client used by a queue. Used when only client specific
 // fields are updated to avoid restarting the queue.
-func (t *QueueManager) SetClient(c StorageClient) {
+func (t *QueueManager) SetClient(c WriteClient) {
 	t.clientMtx.Lock()
 	t.storeClient = c
 	t.clientMtx.Unlock()
 }
 
-func (t *QueueManager) client() StorageClient {
+func (t *QueueManager) client() WriteClient {
 	t.clientMtx.RLock()
 	defer t.clientMtx.RUnlock()
 	return t.storeClient
 }
 
-func internLabels(lbls labels.Labels) {
+func (t *QueueManager) internLabels(lbls labels.Labels) {
 	for i, l := range lbls {
-		lbls[i].Name = interner.intern(l.Name)
-		lbls[i].Value = interner.intern(l.Value)
+		lbls[i].Name = t.interner.intern(l.Name)
+		lbls[i].Value = t.interner.intern(l.Value)
 	}
 }
 
-func releaseLabels(ls labels.Labels) {
+func (t *QueueManager) releaseLabels(ls labels.Labels) {
 	for _, l := range ls {
-		interner.release(l.Name)
-		interner.release(l.Value)
+		t.interner.release(l.Name)
+		t.interner.release(l.Value)
 	}
 }
 
@@ -538,7 +673,7 @@ func (t *QueueManager) shouldReshard(desiredShards int) bool {
 	// We shouldn't reshard if Prometheus hasn't been able to send to the
 	// remote endpoint successfully within some period of time.
 	minSendTimestamp := time.Now().Add(-2 * time.Duration(t.cfg.BatchSendDeadline)).Unix()
-	lsts := atomic.LoadInt64(&t.lastSendTimestamp)
+	lsts := t.lastSendTimestamp.Load()
 	if lsts < minSendTimestamp {
 		level.Warn(t.logger).Log("msg", "Skipping resharding, last successful send was beyond threshold", "lastSendTimestamp", lsts, "minSendTimestamp", minSendTimestamp)
 		return false
@@ -566,7 +701,7 @@ func (t *QueueManager) calculateDesiredShards() int {
 		samplesOutDuration = t.samplesOutDuration.rate() / float64(time.Second)
 		samplesPendingRate = samplesInRate*samplesKeptRatio - samplesOutRate
 		highestSent        = t.metrics.highestSentTimestamp.Get()
-		highestRecv        = highestTimestamp.Get()
+		highestRecv        = t.highestRecvTimestamp.Get()
 		delay              = highestRecv - highestSent
 		samplesPending     = delay * samplesInRate * samplesKeptRatio
 	)
@@ -664,14 +799,15 @@ type shards struct {
 	// Emulate a wait group with a channel and an atomic int, as you
 	// cannot select on a wait group.
 	done    chan struct{}
-	running int32
+	running atomic.Int32
 
 	// Soft shutdown context will prevent new enqueues and deadlocks.
 	softShutdown chan struct{}
 
 	// Hard shutdown context is used to terminate outgoing HTTP connections
 	// after giving them a chance to terminate.
-	hardShutdown context.CancelFunc
+	hardShutdown          context.CancelFunc
+	droppedOnHardShutdown atomic.Uint32
 }
 
 // start the shards; must be called before any call to enqueue.
@@ -679,6 +815,9 @@ func (s *shards) start(n int) {
 	s.mtx.Lock()
 	defer s.mtx.Unlock()
 
+	s.qm.metrics.pendingSamples.Set(0)
+	s.qm.metrics.numShards.Set(float64(n))
+
 	newQueues := make([]chan sample, n)
 	for i := 0; i < n; i++ {
 		newQueues[i] = make(chan sample, s.qm.cfg.Capacity)
@@ -689,12 +828,12 @@ func (s *shards) start(n int) {
 	var hardShutdownCtx context.Context
 	hardShutdownCtx, s.hardShutdown = context.WithCancel(context.Background())
 	s.softShutdown = make(chan struct{})
-	s.running = int32(n)
+	s.running.Store(int32(n))
 	s.done = make(chan struct{})
+	s.droppedOnHardShutdown.Store(0)
 	for i := 0; i < n; i++ {
 		go s.runShard(hardShutdownCtx, i, newQueues[i])
 	}
-	s.qm.metrics.numShards.Set(float64(n))
 }
 
 // stop the shards; subsequent call to enqueue will return false.
@@ -719,12 +858,14 @@ func (s *shards) stop() {
 	case <-s.done:
 		return
 	case <-time.After(s.qm.flushDeadline):
-		level.Error(s.qm.logger).Log("msg", "Failed to flush all samples on shutdown")
 	}
 
 	// Force an unclean shutdown.
 	s.hardShutdown()
 	<-s.done
+	if dropped := s.droppedOnHardShutdown.Load(); dropped > 0 {
+		level.Error(s.qm.logger).Log("msg", "Failed to flush all samples on shutdown", "count", dropped)
+	}
 }
 
 // enqueue a sample.  If we are currently in the process of shutting down or resharding,
@@ -744,13 +885,14 @@ func (s *shards) enqueue(ref uint64, sample sample) bool {
 	case <-s.softShutdown:
 		return false
 	case s.queues[shard] <- sample:
+		s.qm.metrics.pendingSamples.Inc()
 		return true
 	}
 }
 
 func (s *shards) runShard(ctx context.Context, shardID int, queue chan sample) {
 	defer func() {
-		if atomic.AddInt32(&s.running, -1) == 0 {
+		if s.running.Dec() == 0 {
 			close(s.done)
 		}
 	}()
@@ -781,6 +923,12 @@ func (s *shards) runShard(ctx context.Context, shardID int, queue chan sample) {
 	for {
 		select {
 		case <-ctx.Done():
+			// In this case we drop all samples in the buffer and the queue.
+			// Remove them from pending and mark them as failed.
+			droppedSamples := nPending + len(queue)
+			s.qm.metrics.pendingSamples.Sub(float64(droppedSamples))
+			s.qm.metrics.failedSamplesTotal.Add(float64(droppedSamples))
+			s.droppedOnHardShutdown.Add(uint32(droppedSamples))
 			return
 
 		case sample, ok := <-queue:
@@ -801,7 +949,6 @@ func (s *shards) runShard(ctx context.Context, shardID int, queue chan sample) {
 			pendingSamples[nPending].Samples[0].Timestamp = sample.t
 			pendingSamples[nPending].Samples[0].Value = sample.v
 			nPending++
-			s.qm.metrics.pendingSamples.Inc()
 
 			if nPending >= max {
 				s.sendSamples(ctx, pendingSamples, &buf)
@@ -836,28 +983,27 @@ func (s *shards) sendSamples(ctx context.Context, samples []prompb.TimeSeries, b
 	// should be maintained irrespective of success or failure.
 	s.qm.samplesOut.incr(int64(len(samples)))
 	s.qm.samplesOutDuration.incr(int64(time.Since(begin)))
-	atomic.StoreInt64(&s.qm.lastSendTimestamp, time.Now().Unix())
+	s.qm.lastSendTimestamp.Store(time.Now().Unix())
 }
 
 // sendSamples to the remote storage with backoff for recoverable errors.
 func (s *shards) sendSamplesWithBackoff(ctx context.Context, samples []prompb.TimeSeries, buf *[]byte) error {
-	req, highest, err := buildWriteRequest(samples, *buf)
+	// Build the WriteRequest with no metadata.
+	req, highest, err := buildWriteRequest(samples, nil, *buf)
 	if err != nil {
 		// Failing to build the write request is non-recoverable, since it will
 		// only error if marshaling the proto to bytes fails.
 		return err
 	}
 
-	backoff := s.qm.cfg.MinBackoff
 	reqSize := len(*buf)
 	sampleCount := len(samples)
 	*buf = req
-	try := 0
 
 	// An anonymous function allows us to defer the completion of our per-try spans
 	// without causing a memory leak, and it has the nice effect of not propagating any
 	// parameters for sendSamplesWithBackoff/3.
-	attemptStore := func() error {
+	attemptStore := func(try int) error {
 		span, ctx := opentracing.StartSpanFromContext(ctx, "Remote Send Batch")
 		defer span.Finish()
 
@@ -868,6 +1014,7 @@ func (s *shards) sendSamplesWithBackoff(ctx context.Context, samples []prompb.Ti
 		span.SetTag("remote_url", s.qm.storeClient.Endpoint())
 
 		begin := time.Now()
+		s.qm.metrics.samplesTotal.Add(float64(sampleCount))
 		err := s.qm.client().Store(ctx, *buf)
 		s.qm.metrics.sentBatchDuration.Observe(time.Since(begin).Seconds())
 
@@ -880,6 +1027,23 @@ func (s *shards) sendSamplesWithBackoff(ctx context.Context, samples []prompb.Ti
 		return nil
 	}
 
+	onRetry := func() {
+		s.qm.metrics.retriedSamplesTotal.Add(float64(sampleCount))
+	}
+
+	err = sendWriteRequestWithBackoff(ctx, s.qm.cfg, s.qm.logger, attemptStore, onRetry)
+	if err != nil {
+		return err
+	}
+	s.qm.metrics.samplesBytesTotal.Add(float64(reqSize))
+	s.qm.metrics.highestSentTimestamp.Set(float64(highest / 1000))
+	return nil
+}
+
+func sendWriteRequestWithBackoff(ctx context.Context, cfg config.QueueConfig, l log.Logger, attempt func(int) error, onRetry func()) error {
+	backoff := cfg.MinBackoff
+	try := 0
+
 	for {
 		select {
 		case <-ctx.Done():
@@ -887,37 +1051,34 @@ func (s *shards) sendSamplesWithBackoff(ctx context.Context, samples []prompb.Ti
 		default:
 		}
 
-		err = attemptStore()
+		err := attempt(try)
 
-		if err != nil {
-			// If the error is unrecoverable, we should not retry.
-			if _, ok := err.(recoverableError); !ok {
-				return err
-			}
-
-			// If we make it this far, we've encountered a recoverable error and will retry.
-			s.qm.metrics.retriedSamplesTotal.Add(float64(sampleCount))
-			level.Warn(s.qm.logger).Log("msg", "Failed to send batch, retrying", "err", err)
-			time.Sleep(time.Duration(backoff))
-			backoff = backoff * 2
-
-			if backoff > s.qm.cfg.MaxBackoff {
-				backoff = s.qm.cfg.MaxBackoff
-			}
-
-			try++
-			continue
+		if err == nil {
+			return nil
 		}
 
-		// Since we retry forever on recoverable errors, this needs to stay inside the loop.
-		s.qm.metrics.succeededSamplesTotal.Add(float64(sampleCount))
-		s.qm.metrics.bytesSent.Add(float64(reqSize))
-		s.qm.metrics.highestSentTimestamp.Set(float64(highest / 1000))
-		return nil
+		// If the error is unrecoverable, we should not retry.
+		if _, ok := err.(RecoverableError); !ok {
+			return err
+		}
+
+		// If we make it this far, we've encountered a recoverable error and will retry.
+		onRetry()
+		level.Warn(l).Log("msg", "Failed to send batch, retrying", "err", err)
+
+		time.Sleep(time.Duration(backoff))
+		backoff = backoff * 2
+
+		if backoff > cfg.MaxBackoff {
+			backoff = cfg.MaxBackoff
+		}
+
+		try++
+		continue
 	}
 }
 
-func buildWriteRequest(samples []prompb.TimeSeries, buf []byte) ([]byte, int64, error) {
+func buildWriteRequest(samples []prompb.TimeSeries, metadata []prompb.MetricMetadata, buf []byte) ([]byte, int64, error) {
 	var highest int64
 	for _, ts := range samples {
 		// At the moment we only ever append a TimeSeries with a single sample in it.
@@ -925,8 +1086,10 @@ func buildWriteRequest(samples []prompb.TimeSeries, buf []byte) ([]byte, int64,
 			highest = ts.Samples[0].Timestamp
 		}
 	}
+
 	req := &prompb.WriteRequest{
 		Timeseries: samples,
+		Metadata:   metadata,
 	}
 
 	data, err := proto.Marshal(req)
